package com.tzld.piaoquan.recommend.framework.merger;

import com.tzld.piaoquan.recommend.framework.common.item.RankerItem;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.framework.common.user.User;
import com.tzld.piaoquan.recommend.framework.retrieve.candidate.Candidate;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;


public abstract class StrategyQueue {
    protected final Map<String, Pair<MergeRule, List<RankerItem>>> rankerItemsList = new HashMap<String, Pair<MergeRule, List<RankerItem>>>();
    private final Logger LOGGER = LoggerFactory.getLogger(StrategyQueue.class);
    protected Map<String, StrategyQueue> children;

    protected List<RankerItem> items;
    private StrategyQueueInfo strategyQueueInfo;
    private List<StrategyQueue> allQueues = null;

    public StrategyQueue(StrategyQueueInfo strategyQueueInfo, StrategyQueueConfig strategyQueueConfig) {
        this.strategyQueueInfo = strategyQueueInfo;

        items = new LinkedList<RankerItem>();
        children = new HashMap<String, StrategyQueue>();

        for (String childName : strategyQueueInfo.getChildren()) {
            if (!strategyQueueConfig.getStrategyQueueInfoMap().containsKey(childName)) {
                LOGGER.error("No config for Child queue [{}]", childName);
                continue;
            }
            StrategyQueueInfo childQueueInfo = strategyQueueConfig.getStrategyQueueInfoMap().get(childName);

            StrategyQueue childQueue = null;
            String childQueueClass = childQueueInfo.getQueueClass();
            try {
                childQueue = (StrategyQueue) Class.forName(childQueueClass).getConstructor(childQueueInfo.getClass(), strategyQueueConfig.getClass()).newInstance(childQueueInfo, strategyQueueConfig);
            } catch (InstantiationException e) {
                LOGGER.error("construct StrategyQueue {}: [{}]", childName + " [" + childQueueClass + "]", e);
            } catch (IllegalAccessException e) {
                LOGGER.error("construct StrategyQueue {}: [{}]", childName + " [" + childQueueClass + "]", e);
            } catch (ClassNotFoundException e) {
                LOGGER.error("construct StrategyQueue {}: [{}]", childName + " [" + childQueueClass + "]", e);
            } catch (NoSuchMethodException e) {
                LOGGER.error("construct StrategyQueue {}: [{}]", childName + " [" + childQueueClass + "]", e);
            } catch (InvocationTargetException e) {
                LOGGER.error("construct StrategyQueue {}: [{}]", childName + " [" + childQueueClass + "]", e);
            }
            if (childQueue != null) {
                putChild(childName, childQueue);
            } else {
                LOGGER.error("construct child queue [{}] failed", childName);
            }
        }

//        getAllQueues();
    }

    public StrategyQueueInfo getStrategyQueueInfo() {
        return strategyQueueInfo;
    }

    public List<StrategyQueue> getAllQueues() {
        if (allQueues == null) {
            allQueues = new LinkedList<StrategyQueue>();

            if (!strategyQueueInfo.isLeaf()) {
                for (Map.Entry<String, StrategyQueue> child : children.entrySet()) {
                    allQueues.addAll(child.getValue().getAllQueues());
                }
            }

            allQueues.add(this);
        }
        return allQueues;
    }

    private void putChild(String queueName, StrategyQueue child) {
        children.put(queueName, child);
    }

    public List<RankerItem> getItems() {
        return items;
    }

    public void setItems(List<RankerItem> items) {
        this.items = items;
    }

    public void clearItems() {
        items.clear();
    }

    public void addItem(RankerItem item) {
        this.items.add(item);
    }

    public abstract int doMerge(final Map<String, Pair<MergeRule, List<RankerItem>>> rankerItemsListMap, final int recNum, final User user, final RecommendRequest requestData, final int requestIndex, final int expId);

    public abstract int candidate(Map<String, Candidate> candidateMap, final int recallNum, final User user, final RecommendRequest requestData, final int requestIndex, final int expId);

    public final int merge(final int recNum, final User user, final RecommendRequest requestData, final int requestIndex, final int expId) {
        this.beforeMerge(recNum, user, requestData, requestIndex, expId);
        this.doMerge(rankerItemsList, recNum, user, requestData, requestIndex, expId);
        this.doDeDup();
        return this.afterMerge();
    }

    public void beforeMerge(final int recNum, final User user, final RecommendRequest requestData, final int requestIndex, final int expId) {

        if (strategyQueueInfo.isLeaf()) {
            LOGGER.debug("leaf queue [{}] with [{}] items to merge", getStrategyQueueInfo().getQueueName(), items.size());
        } else {
            rankerItemsList.clear();

            // XXX: what if a queue recall items by itself?
            clearItems();

            int queueCount = 0, itemCount = 0;
            for (MergeRule rule : strategyQueueInfo.getRulesList()) {
                if (rule.isDisabled(expId)) {
                    continue;
                }

                int myRecNum = Math.min(recNum, rule.maxMergeNum);
                int actualRecNum = children.get(rule.queueName).merge(myRecNum, user, requestData, requestIndex, expId);

                queueCount += 1;
                itemCount += actualRecNum;

                rankerItemsList.put(rule.queueName, Pair.of(rule, children.get(rule.queueName).getItems()));
            }

            LOGGER.debug(getStrategyQueueInfo().getQueueName() + " merge info: [{}] queues with [{}] items to merge", queueCount, itemCount);
        }
    }

    /**
     * id dedup
     *
     * @return
     */
    public final int doDeDup() {
        List<RankerItem> pureItems = new ArrayList<RankerItem>();
        Set<String> deDupItems = new HashSet<String>();
        int deDupCount = 0;
        for (RankerItem item : items) {
            if (deDupItems.contains(item.getId())) {
                if (LOGGER.isDebugEnabled()) {
                    LOGGER.debug("merge queue dedup item, queue [{}] itemid [{}]",
                            strategyQueueInfo.getQueueName(), item.getId());
                }
                deDupCount++;
                continue;
            }

            deDupItems.add(item.getId());
            pureItems.add(item);
        }
        items = pureItems;
        LOGGER.debug("dedup items, queue [{}], dedup number [{}], current number [{}]",
                new Object[]{getStrategyQueueInfo().getQueueName(), deDupCount, items.size()});
        return items.size();
    }

    public int afterMerge() {
        int n = items.size();
        for (RankerItem item : items) {
            item.addMergeQueuePath(getStrategyQueueInfo().getQueueName());
        }
        LOGGER.debug("number of items after [{}] merge: [{}]", getStrategyQueueInfo().getQueueName(), n);
        return n;
    }

    protected int simpleCandidateByPercentage(Map<String, Candidate> candidateMap, int recallNum, User user, RecommendRequest requestData, int requestIndex, int expId) {
        int n = 0;

        for (MergeRule rule : strategyQueueInfo.getRulesList()) {
            if (rule.isDisabled(expId)) {
                continue;
            }

            int myRecallNum = (int) (recallNum * rule.recallPercentage);
            if (!children.containsKey(rule.queueName)) {
                LOGGER.error("Rule for not exist child queue [{}]", rule.queueName);
            }

            Map<String, Candidate> myCandidates = new HashMap<String, Candidate>();
            n += children.get(rule.queueName).candidate(myCandidates, myRecallNum, user, requestData, requestIndex, expId);
            candidateMap.putAll(myCandidates);
        }

        return n;
    }

}
