package com.tzld.piaoquan.recommend.framework.recaller;


import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.base.Optional;
import com.google.common.base.Predicate;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import com.tzld.piaoquan.recommend.framework.common.item.ItemProvider;
import com.tzld.piaoquan.recommend.framework.common.item.RankerItem;
import com.tzld.piaoquan.recommend.framework.retrieve.candidate.*;
import com.tzld.piaoquan.recommend.framework.common.user.User;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;


import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;


public abstract class AbstractRecaller<T> {


    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractRecaller.class);
    private static final long DEFAULT_QUEUE_LOAD_TIMEOUT = 150; // ms
    private static final long DEFAULT_PARALLEL_FILTER_TIMEOUT = 200; // ms
    private static final ExecutorService filterExecutorService = Executors.newFixedThreadPool(128);
    private static final ExecutorService fetchQueueExecutorService = Executors.newFixedThreadPool(128);

    protected final ItemProvider<T> itemProvider;
    private final QueueProvider<T> queueProvider;
    private final FilterConfig filterConfig;
    private final long QUEUE_LOAD_TIMEOUT;

    public AbstractRecaller(ItemProvider<T> itemProvider, QueueProvider<T> queueProvider, FilterConfig filterConfig) {
        this(itemProvider, queueProvider, filterConfig, DEFAULT_QUEUE_LOAD_TIMEOUT);
    }

    public AbstractRecaller(ItemProvider<T> itemProvider, QueueProvider<T> queueProvider,
                            FilterConfig filterConfig, long queueLoadTimeout) {
        this.itemProvider = itemProvider;
        this.queueProvider = queueProvider;
        this.filterConfig = filterConfig;
        this.QUEUE_LOAD_TIMEOUT = queueLoadTimeout;
    }

    public abstract Map<String, Long> getCacheRulesConfig();

    public abstract String extractItemId(Entry<T> entry);

    public abstract Object extractItemBytes(Entry<T> entry);
    public abstract boolean isValidItem(T t);

    public abstract List<Candidate> filterCandidates(RecommendRequest requestData, User user, int requestIndex, List<Candidate> candidates);

    public ItemProvider<T> getItemProvider() {
        return itemProvider;
    }

    public QueueProvider<T> getQueueProvider() {
        return queueProvider;
    }

    public Map<Candidate, Queue<T>> getQueue(String queueStr) throws Exception {
        Candidate candidate = new Candidate();
        candidate.setCandidateKey(queueStr);
        return loadQueues(Arrays.asList(candidate));
    }

    public Optional<T> getItem(String itemId) throws Exception {
        return itemProvider.get(itemId);
    }

    public long getItemDBSize() {
        return this.itemProvider.dbSize();
    }

    public long getIndexDBSize() {
        return this.queueProvider.dbSize();
    }

    public long getQueueTTL(String queueNameStr, Map<String, Long> cacheRules) {

//        LOGGER.info("[recall server] queue name from string={}", queueNameStr);
        QueueName name = QueueName.fromString(queueNameStr);

        for (String match : name.getMatches()) {
            if (cacheRules.containsKey(match)) {
                return cacheRules.get(match);
            }
        }
        return name.getTTL();
    }

    private Map<String, Double> listToMap(List<String> list) {
        Map<String, Double> map = new HashMap<String, Double>();
        if (list != null) {
            for (String elem : list) {
                map.put(elem, 1.0);
            }
        }
        return map;
    }

    /**
     * @param entries
     * @param candidate
     * @param user
     * @return
     */
    private List<RankerItem> toHits(final Iterable<Entry<T>> entries, final Candidate candidate, final User user) {

        List<RankerItem> result = new ArrayList<RankerItem>();
        for (Entry entry : entries) {
            RankerItem item = new RankerItem();
            item.setId(extractItemId(entry));

            CandidateInfo candidateInfo = new CandidateInfo();
            candidateInfo.setCandidate_queue(candidate.getCandidateKey());
            candidateInfo.setCandidate(candidate);
            item.setCandidateInfo(candidateInfo);
            item.addToCandidateInfoList(candidateInfo);

            result.add(item);
        }
        return result;
    }

    public Map<Candidate, Queue<T>> loadQueues(List<Candidate> candidates) {

        // update queueName
        final Map<String, Long> cacheRules = getCacheRulesConfig();
        Iterable<Candidate> updateCandidates = FluentIterable.from(candidates).transform(new Function<Candidate, Candidate>() {
            @Override
            public Candidate apply(Candidate candidate) {
                try {
                    long ttl = getQueueTTL(candidate.getCandidateKey(), cacheRules);
                    candidate.setCandidateQueueName(QueueName.fromString(candidate.getCandidateKey(), ttl));
                } catch (Exception e) {
                    candidate.setCandidateQueueName(null);
                    LOGGER.error("error parse QueueName [{}]", candidate.getCandidateKey());
                }
                return candidate;
            }
        }).filter(new Predicate<Candidate>() {
            @Override
            public boolean apply(Candidate candidate) {
                return candidate.getCandidateQueueName() != null && !StringUtils.equals("dssm-exploit-index", candidate.getQueue())
                        && !StringUtils.equals("dnnytb-exploit-index", candidate.getQueue());
            }
        });

        // parse queues
        Iterable<QueueName> queueNames = FluentIterable.from(updateCandidates).transform(new Function<Candidate, QueueName>() {
            @Override
            public QueueName apply(Candidate candidate) {
                return candidate.getCandidateQueueName();
            }
        });

        // parallel load queues
        Map<QueueName, Queue<T>> queues = Maps.newConcurrentMap();
        try {
            queues = queueProvider.loads(Lists.newArrayList(queueNames), QUEUE_LOAD_TIMEOUT, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            LOGGER.error("load queue occur error [{}]", ExceptionUtils.getFullStackTrace(e));
        }

        // parse candidate map
        Map<Candidate, Queue<T>> candidateQueueMap = Maps.newConcurrentMap();
        for (Candidate candidate : updateCandidates) {
            QueueName name = candidate.getCandidateQueueName();
            if (queues.containsKey(name) && queues.get(name) != null) {
                candidateQueueMap.put(candidate, queues.get(name));
            }
        }
        return candidateQueueMap;
    }


    /**
     * recall
     * @param requestData
     * @param user
     * @param requestIndex
     * @param recallCandidates
     * @return
     */
    public List<RankerItem> recalling(final RecommendRequest requestData, final User user, int requestIndex, List<Candidate> recallCandidates) {

        long startTime = System.currentTimeMillis();
        final RecallFilter<T> recallFilter = new RecallFilter<T>(this.filterConfig, requestData, user, requestIndex);

        // filter candidates
        final List<Candidate> refactorCandidates = filterCandidates(requestData, user, requestIndex, recallCandidates);

        // load queue
        List<Callable<Map<Candidate, Queue<T>>>> fetchQueueCalls = Lists.newArrayList();

        fetchQueueCalls.add(new Callable<Map<Candidate, Queue<T>>>() {
            @Override
            public Map<Candidate, Queue<T>> call() throws Exception {
                boolean isFromRedis = true;
                return obtainQueue(refactorCandidates, requestData, user, isFromRedis);
            }
        });

        fetchQueueCalls.add(new Callable<Map<Candidate, Queue<T>>>() {
            @Override
            public Map<Candidate, Queue<T>> call() throws Exception {
                boolean isFromRedis = false;
                return obtainQueue(refactorCandidates, requestData, user, isFromRedis);
            }
        });

        List<Future<Map<Candidate, Queue<T>>>> fetchQueueFutures = null;

        try {
            fetchQueueFutures = fetchQueueExecutorService.invokeAll(fetchQueueCalls, DEFAULT_QUEUE_LOAD_TIMEOUT,
                    TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            LOGGER.error("[fetch queue error] inter fail: {}", ExceptionUtils.getStackTrace(e));
        } catch (Exception e) {
            LOGGER.error("[fetch queue error] ex", ExceptionUtils.getStackTrace(e));
        }

        Map<Candidate, Queue<T>> candidateQueueMap = Maps.newHashMap();
        if (CollectionUtils.isNotEmpty(fetchQueueFutures)) {

            for (Future<Map<Candidate, Queue<T>>> future : fetchQueueFutures) {

                if (future.isDone() && !future.isCancelled()) {

                    Map<Candidate, Queue<T>> result = null;
                    try {
                        result = future.get();
                    } catch (InterruptedException e) {
                        LOGGER.error("[fetch queue error] InterruptedException {}", ExceptionUtils.getStackTrace(e));
                    } catch (ExecutionException e) {
                        LOGGER.error("[fetch queue error] ex {}", ExceptionUtils.getStackTrace(e));
                    }

                    if (result != null) {
                        candidateQueueMap.putAll(result);
                    }
                }
            }

        }

        // do filter
        long filterStartTime = System.currentTimeMillis();

        List<Map.Entry<Candidate, Queue<T>>> batch = new ArrayList<Map.Entry<Candidate, Queue<T>>>();
        final List<Callable<List<RankerItem>>> callables = new ArrayList<Callable<List<RankerItem>>>();
        int expectedRecallSum = 0;
        for(final Map.Entry<Candidate, Queue<T>> entry : candidateQueueMap.entrySet()) {
            callables.add(new Callable<List<RankerItem>>() {
                @Override
                public List<RankerItem> call() throws Exception {
                    List<RankerItem> candidateHits = new ArrayList<RankerItem>();
                    final Candidate candidate = entry.getKey();
                    try {
                        // 1. filter
                        Iterable<Entry<T>> entries = FluentIterable.from(entry.getValue()).filter(new Predicate<Entry<T>>() {
                            @Override
                            public boolean apply(Entry<T> entry) {
                                return isValidItem(entry.item) &&
                                        recallFilter.predicate(candidate, entry.item);
                            }
                        }).limit(candidate.getCandidateNum());

                        // 2. toHits
                        candidateHits.addAll(toHits(entries, candidate, user));

                        // debug log for tracing
                        LOGGER.debug("recalled candidate [{}], queue length [{}], expected [{}], hit [{}]",
                                new Object[]{candidate.getCandidateKey(), entry.getValue().size(), candidate.getCandidateNum(), candidateHits.size()});
                    } catch (Exception e) {
                        LOGGER.error("recall filter queue occur error, queue [{}], error: [{}]", candidate.toString(), ExceptionUtils.getFullStackTrace(e));
                    }

                    return candidateHits;
                }
            });
        }

        Map<String, RankerItem> hits = new HashMap<String, RankerItem>();
        try {
            List<Future<List<RankerItem>>> futures = filterExecutorService.invokeAll(callables, DEFAULT_PARALLEL_FILTER_TIMEOUT, TimeUnit.MILLISECONDS);

            for (Future<List<RankerItem>> future : futures) {
                try {
                    if (future.isDone() && !future.isCancelled() && future.get() != null) {
                        List<RankerItem> part = future.get();
                        // add to result
                        for (RankerItem item : part) {
                            // merge candidate Info
                            if (hits.containsKey(item.getId())) {
                                hits.get(item.getId()).addToCandidateInfoList(item.getCandidateInfo());
                            } else {
                                hits.put(item.getId(), item);
                            }
                        }
                        // log
                    } else {

                    }
                } catch (Exception e) {
                  //
                }
            }
        } catch (Exception e) {
            //
        }

        List<RankerItem> result = new ArrayList<RankerItem>(hits.values());
        return result;
    }

    private Map<Candidate, Queue<T>> obtainQueue(List<Candidate> refactorCandidates, RecommendRequest requestData, User user, boolean isFromRedis) {

        if (isFromRedis) {
            return loadQueues(refactorCandidates);
        } else {
            return fetchQueues(refactorCandidates, requestData, user);
        }

    }

    protected Map<Candidate, Queue<T>> fetchQueues(List<Candidate> candidates, RecommendRequest requestData, User user) {

        return null;
    }
}
