package com.tzld.piaoquan.recommend.framework.score;


import com.xiaomi.common.perfcounter.PerfCounter;
import com.xiaomi.data.recommend.common.PerfCounterUtils;
import com.xiaomi.data.recommend.common.RecommendRequestData;
import com.xiaomi.data.recommend.common.model.RankerItem;
import com.xiaomi.data.recommend.common.model.User;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

public class ScorerPipeline {
    public static final int corePoolSize = 128;
    public static final int SCORE_TIME_OUT = 400;
    public static final Logger LOGGER = LoggerFactory.getLogger(ScorerPipeline.class);
    public static final ExecutorService executorService = Executors.newFixedThreadPool(corePoolSize);

    public List<AbstractScorer> scorers;

    public ScorerPipeline(List<AbstractScorer> scorers) {
        this.scorers = scorers;
    }

    /**
     * scoring
     *
     * @param requestContext
     * @param user
     * @param requestIndex
     * @param rankerItems
     * @return
     */
    public List<RankerItem> scoring(final RecommendRequestData requestContext,
                                    final User user,
                                    final int requestIndex,
                                    final List<RankerItem> rankerItems) {
        // check
        if (CollectionUtils.isEmpty(rankerItems)) {
            return rankerItems;
        }
        if (CollectionUtils.isEmpty(scorers)) {
            PerfCounter.count(PerfCounterUtils.SCORER_SCORERS_EMPTY_COUNT, 1);
        }
        long scoreStart = System.currentTimeMillis();
        List<RankerItem> items = rankerItems;
        for (final AbstractScorer scorer : scorers) {

            if (!scorer.isEnable()) {
                continue;
            }

            final int beforeSize = items.size();
            final long startTime = System.currentTimeMillis();

            String fullScorerName = scorer.getScorerConfigInfo().getScorerName();
            String[] scorerNames = fullScorerName.split("\\.");
            final String scorerName = scorerNames.length > 0 ? scorerNames[scorerNames.length - 1] : fullScorerName;

            final List<RankerItem> scoreRankerItems = items;
            Callable<List<RankerItem>> callable = new Callable<List<RankerItem>>() {
                @Override
                public List<RankerItem> call() throws Exception {
                    return scorer.scoring(requestContext, user, requestIndex, scoreRankerItems);
                }
            };

            // execute score use thread to protected score worst time
            List<RankerItem> scoredItems = new ArrayList<RankerItem>();
            try {
                List<Future<List<RankerItem>>> futures = executorService.invokeAll(Arrays.asList(callable), SCORE_TIME_OUT, TimeUnit.MILLISECONDS);
                for (Future<List<RankerItem>> future : futures) {
                    try {
                        if (future.isDone() && !future.isCancelled() && future.get() != null) {
                            scoredItems.addAll(future.get());
                        } else {
                            PerfCounter.count("feeds.scorer.item.fail.number" + scorerName, scoreRankerItems.size());
                            LOGGER.error("score task is cancelled,  uid [{}] scorename [{}] fail items [{}]",
                                    new Object[]{user.getId(), scorerName, scoreRankerItems.size()});
                        }
                    } catch (Exception e) {
                        PerfCounter.count("feeds.scorer.item.fail.number" + scorerName, scoreRankerItems.size());
                        LOGGER.error("thread pool exception uid [{}] scorename [{}], exception [{}]",
                                new Object[]{user.getId(), scorerName, ExceptionUtils.getFullStackTrace(e)});
                    }
                }
            } catch (Exception e) {
                PerfCounter.count("feeds.scorer.item.fail.number" + scorerName, scoreRankerItems.size());
                LOGGER.error("thread pool exception uid [{}] scorename [{}], exception [{}]",
                        new Object[]{user.getId(), scorerName, ExceptionUtils.getFullStackTrace(e)});
            }

            // 变更item
            if (CollectionUtils.isNotEmpty(scoredItems)) {
                items = scoredItems;
            } else {
                items = new ArrayList<>(items);
            }

            int position = 0;
            for (RankerItem item : items) {
                item.putRankerIndex(scorerName, position++);
                item.putRankerScore(scorerName, item.getScore());
            }
            long spentTime = System.currentTimeMillis() - startTime;
            PerfCounter.setHistogramValue("feeds.scorer.item.numbers." + scorerName, beforeSize);
            PerfCounter.setHistogramValue("feeds.scorer.duration." + scorerName, spentTime);
            LOGGER.debug("after scorer [{}], spentTime [{}], before size [{}], remaining size [{}]",
                    new Object[]{scorerName, spentTime, beforeSize, items.size()});
        }

        int position = 0;
        for (RankerItem item : items) {
            item.putRankerIndex("finalScore", position++);
            item.putRankerScore("finalScore", item.getScore());
        }
        PerfCounter.setHistogramValue(PerfCounterUtils.SCORER_OVERALL_DURATION, System.currentTimeMillis() - scoreStart);
        return items;
    }
}
