package com.tzld.piaoquan.recommend.framework.userattention;

import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigObject;
import com.typesafe.config.ConfigValue;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Copyright (c) 2018 XiaoMi Inc. All Rights Reserved.
 * Authors: Liu Jianquan <liujianquan@xiaomi.com>
 */

public class UserAttentionExtractorUtils {

    private static final Logger LOGGER = LoggerFactory.getLogger(UserAttentionExtractorUtils.class);

    public static void initLoadModel(Config config) {
        initLoadModel(constructExtractorList(constructExtractorConfigList(config)));
    }

    public static UserAttentionExtractorPipeline getAttentionPipeline(Config baselineConfig, List<Config> configList) {
        return new UserAttentionExtractorPipeline(constructExtractorList(constructExtractorConfigList(baselineConfig, configList)));
    }

    public static UserAttentionExtractorPipeline getAttentionPipeline(Config mergeConfig) {
        return new UserAttentionExtractorPipeline(constructExtractorList(constructExtractorConfigList(mergeConfig)));
    }

    public static List<UserAttentionExtractorConfig> constructExtractorConfigList(Config baseConfig, List<Config> expConfigList) {
        Config mergedConfig = mergeConfig(baseConfig, expConfigList);
        return constructExtractorConfigList(mergedConfig);
    }

    public static List<UserAttentionExtractorConfig> constructExtractorConfigList(Config config) {
        long startTime = System.currentTimeMillis();
        List<UserAttentionExtractorConfig> configList = new ArrayList<UserAttentionExtractorConfig>();
        try {
            Config mergedConfig = config.getConfig("attention-extractor-config");
            for (ConfigObject.Entry<String, ConfigValue> it : mergedConfig.root().entrySet()) {
                LOGGER.debug("deal with base config, attention key: [{}]", it.getKey());
                Config entryConf = ((ConfigObject) it.getValue()).toConfig();
                UserAttentionExtractorConfig extractorConfig = parseExtractorConfig(entryConf);
                addConfigByPriority(configList, extractorConfig);
            }
            // debug log
            if (LOGGER.isDebugEnabled()) {
                for (UserAttentionExtractorConfig extractorConfig : configList) {
                    LOGGER.debug("attention extractor config [{}] [{}]", extractorConfig.getName(), extractorConfig.toString());
                }
            }
        } catch (Exception e) {
            LOGGER.error("construct extractor config {} failed {}", config.toString(), ExceptionUtils.getFullStackTrace(e));
        }
        return configList;
    }

    public static Config getAttentionMergeConfig(Config baseConfig, List<Config> expConfigList) {
        return mergeConfig(baseConfig, expConfigList);
    }

    private static List<AbstractUserAttentionExtractor> constructExtractorList(List<UserAttentionExtractorConfig> configList) {
        long startTime = System.currentTimeMillis();
        List<AbstractUserAttentionExtractor> extractorList = new ArrayList<AbstractUserAttentionExtractor>();
        for (UserAttentionExtractorConfig config : configList) {
            try {
                AbstractUserAttentionExtractor extractor = (AbstractUserAttentionExtractor) Class.forName(config.getName())
                        .getConstructor(UserAttentionExtractorConfig.class)
                        .newInstance(config);
                extractorList.add(extractor);
            } catch (Exception e) {
                LOGGER.error("instance extractor {} failed {}", config.getName(), ExceptionUtils.getFullStackTrace(e));
            }
        }
        return extractorList;
    }

    private static String loadOptionStringConfig(Config config, String path) {
        return config.hasPath(path) ? config.getString(path) : null;
    }

    private static Config loadOptionConfig(Config config, String path) {
        return config.hasPath(path) ? config.getConfig(path) : ConfigFactory.empty();
    }

    private static int loadOptionIntConfig(Config config, String path) {
        return config.hasPath(path) ? config.getInt(path) : 0;
    }

    private static boolean loadOptionBooleanConfig(Config config, String path) {
        return config.hasPath(path) ? config.getBoolean(path) : false;
    }

    private static UserAttentionExtractorConfig parseExtractorConfig(Config extractorConfig) {
        String attentionExtractorName = extractorConfig.getString("name");
        Integer priority = loadOptionIntConfig(extractorConfig, "attention-priority");
        Boolean disableSwitch = loadOptionBooleanConfig(extractorConfig, "disable-switch");
        Config paramConfig = loadOptionConfig(extractorConfig, "param-config");
        String modelPath = loadOptionStringConfig(extractorConfig, "model-path");
        return new UserAttentionExtractorConfig(attentionExtractorName,
                paramConfig,
                priority,
                disableSwitch,
                modelPath);
    }

    private static Config mergeConfig(Config baselineConfig, List<Config> expConfigs) {
        if (expConfigs == null || expConfigs.size() == 0) {
            return baselineConfig;
        }
        Config resultConfig = expConfigs.get(0);
        for (Config config : expConfigs) {
            resultConfig = resultConfig.withFallback(config);
        }
        resultConfig = resultConfig.withFallback(baselineConfig);
        return resultConfig;
    }

    /**
     * init load model
     *
     * @param attentionExtractors
     */
    private static void initLoadModel(List<AbstractUserAttentionExtractor> attentionExtractors) {
        for (AbstractUserAttentionExtractor attentionExtractor : attentionExtractors) {
            if (attentionExtractor.isEnable())
                attentionExtractor.loadModel();
        }
    }

    private static void addConfigByPriority(List<UserAttentionExtractorConfig> configInfoList, UserAttentionExtractorConfig extractorConfig) {

        int pos = 0;
        for (; pos < configInfoList.size(); pos++) {
            if (configInfoList.get(pos).getPriority() <= extractorConfig.getPriority()) {
                break;
            }
        }
        configInfoList.add(pos, extractorConfig);
    }
}

